/* Routines for dyadic grid plots, general dimension. */
/* Last edited on 2007-01-04 00:13:35 by stolfi */

#ifndef plotnd_H
#define plotnd_H

#include <stdio.h>

#include <pswr.h>

#include <dg_grid.h>
#include <dg_locus.h>
#include <dg_pulse.h>
#include <dg_tent.h>
#include <dg_tree.h>

#include <figtools.h>

/* ELEMENT RANGE ESTIMATION */

void compute_nd_tent_ranges
  ( dg_dim_t d,              /* Domain dimension. */
    dg_pulse_family_t fam[], 
    dg_pulse_t p[],
    double wd[],             /* Widths of root cell. */
    interval_t fr[]          /* (Out) X-, Y-, and value ranges (rounded). */
  );
  /* Computes the nominal domain ranges {fr[0..d-1]} and value range
    {fr[d]} that are suitable for plotting a standard dyadic tent
    function that is the product of pulses {p[i]} from family
    {fam[i]}, for {i} in {0..d-1}. assumes that the total grid extent
    along each axis {i} is {wd[i]}. */

void compute_nd_common_tent_ranges
  ( dg_dim_t d,              /* Domain dimension. */
    dg_pulse_kind_t pkind[],  /* Pulse kinds. */
    dg_cont_t c[],            /* Continuity classes. */
    dg_degree_t gmax[],       /* Max degrees. */
    dg_grid_size_t gsz[],     /* Cell counts in grid period. */
    double wd[],              /* Widths of root cell. */
    interval_t fr[]           /* (Out) X-, Y-, and value ranges (rounded). */
  );
  /* Computes the nominal domain ranges {fr[0..d-1]} and value range
    {fr[d]} that are suitable for plotting any standard dyadic tent
    that is the product of pulses of kind {pkind[i]}, continuity
    {c[i]}, any degree up to {gmax[i]}, and any index, on a periodic
    grid with {gsz[i]} cells and extent {wd[i]} along each axis
    {i}. */

#endif
